"use strict";
// Based on https://github.com/yammer/circuit-breaker-js
Object.defineProperty(exports, "__esModule", { value: true });
function noop() { }
class CircuitBreaker {
    constructor(options) {
        options = options || {};
        this.windowDuration = options.windowDuration || 10000;
        this.timeoutDuration = options.timeoutDuration || 3000;
        this.errorThreshold = options.errorThreshold || 50;
        this.volumeThreshold = options.volumeThreshold || 5;
        this.onCircuitOpen = options.onCircuitOpen || noop;
        this.onCircuitClose = options.onCircuitClose || noop;
        this.buckets = [];
        const numberOfBuckets = Math.max(1, options.numBuckets || 10);
        for (let i = 0; i < numberOfBuckets; ++i) {
            this.buckets.push({
                failures: 0,
                successes: 0,
                timeouts: 0,
                shortCircuits: 0
            });
        }
        this.bucketIndex = 0;
        this.state = 2 /* CLOSED */;
        this.forced = undefined;
        this.startTicker();
    }
    run(command, fallback) {
        if (this.isOpen()) {
            this.executeFallback(fallback || function () { });
        }
        else {
            this.executeCommand(command);
        }
    }
    forceClose() {
        this.forced = this.state;
        this.state = 2 /* CLOSED */;
    }
    forceOpen() {
        this.forced = this.state;
        this.state = 0 /* OPEN */;
    }
    unforce() {
        if (this.forced !== undefined) {
            this.state = this.forced;
            this.forced = undefined;
        }
    }
    isOpen() {
        return this.state === 0 /* OPEN */;
    }
    startTicker() {
        const self = this;
        const bucketDuration = this.windowDuration / this.buckets.length;
        const tick = () => {
            ++this.bucketIndex;
            // FIXME this is very broken as it means that the time
            //       till CB is changing to half-open state depends on
            //       the index of the bucket at the time when it opened.
            if (this.bucketIndex >= this.buckets.length) {
                this.bucketIndex = 0;
                if (self.isOpen()) {
                    self.state = 1 /* HALF_OPEN */;
                }
            }
            // Since we are recycling the buckets they need to be
            // reset before the can be used again.
            const bucket = this.lastBucket();
            bucket.failures = 0;
            bucket.successes = 0;
            bucket.timeouts = 0;
            bucket.shortCircuits = 0;
        };
        setInterval(tick, bucketDuration).unref();
    }
    lastBucket() {
        return this.buckets[this.bucketIndex];
    }
    executeCommand(command) {
        const self = this;
        let timeout;
        const increment = function (prop) {
            return function () {
                if (!timeout) {
                    return;
                }
                const bucket = self.lastBucket();
                bucket[prop]++;
                if (self.forced == null) {
                    self.updateState();
                }
                clearTimeout(timeout);
                timeout = undefined;
            };
        };
        timeout = setTimeout(increment("timeouts"), this.timeoutDuration);
        command(increment("successes"), increment("failures"));
    }
    executeFallback(fallback) {
        fallback();
        const bucket = this.lastBucket();
        bucket.shortCircuits++;
    }
    calculateMetrics() {
        let totalCount = 0;
        let errorCount = 0;
        for (const bucket of this.buckets) {
            const errors = bucket.failures + bucket.timeouts;
            errorCount += errors;
            totalCount += errors + bucket.successes;
        }
        const errorPercentage = (errorCount / (totalCount > 0 ? totalCount : 1)) * 100;
        return {
            totalCount,
            errorCount,
            errorPercentage
        };
    }
    updateState() {
        const metrics = this.calculateMetrics();
        if (this.state == 1 /* HALF_OPEN */) {
            const lastCommandFailed = !this.lastBucket().successes && metrics.errorCount > 0;
            if (lastCommandFailed) {
                this.state = 0 /* OPEN */;
            }
            else {
                this.state = 2 /* CLOSED */;
                this.onCircuitClose(metrics);
            }
        }
        else {
            const overErrorThreshold = metrics.errorPercentage > this.errorThreshold;
            const overVolumeThreshold = metrics.totalCount > this.volumeThreshold;
            const overThreshold = overVolumeThreshold && overErrorThreshold;
            if (overThreshold) {
                this.state = 0 /* OPEN */;
                this.onCircuitOpen(metrics);
            }
        }
    }
}
exports.CircuitBreaker = CircuitBreaker;
//# sourceMappingURL=circuit-breaker.js.map