import { CircuitBreaker, CircuitBreakerOptions, Metrics as CircuitBreakerMetrics, CircuitBreakerPublicApi } from "./circuit-breaker";
import { RequestError, ServiceClientRequestOptions, ServiceClientResponse, TimingPhases, Timings } from "./request";
export { CircuitBreaker, CircuitBreakerOptions, CircuitBreakerMetrics, CircuitBreakerPublicApi, ServiceClientResponse, ServiceClientRequestOptions };
/**
 * This interface defines factory function for getting a circuit breaker
 */
export declare type CircuitBreakerFactory = (params: ServiceClientRequestOptions) => CircuitBreaker;
/**
 * A request filter may introduce one or both functions in this interface. For more information
 * regarding request filters, refer to the readme of this project.
 */
export interface ServiceClientRequestFilter {
    /**
     * This callback is called before the requests is done.
     * You can short-circuit the request by both returning
     * a ServiceClient.Response object which is helpful for
     * implementing caching or mocking. You could also
     * fail the request by throwing an Error.
     * @throws {Error}
     */
    request?: (requestOptions: ServiceClientRequestOptions) => ServiceClientResponse | ServiceClientRequestOptions | Promise<ServiceClientResponse | ServiceClientRequestOptions>;
    /**
     * This callback is called after the response has arrived.
     * @throws {Error}
     */
    response?: (response: ServiceClientResponse) => ServiceClientResponse | Promise<ServiceClientResponse>;
}
/**
 * This interface describes all the options that may be passed to the service client at construction time.
 */
export declare class ServiceClientOptions {
    /**
     * This is the only mandatory option when creating a service client. All other properties have
     * reasonable defaults.
     */
    hostname: string;
    /**
     * Name of the client. Primarily used in errors. Defaults to hostname.
     */
    name?: string;
    /**
     * If this property is not provided, the {@link ServiceClient.DEFAULT_FILTERS} will be used.
     */
    filters?: ServiceClientRequestFilter[];
    /**
     * should the service client record request timings?
     */
    timing?: boolean;
    autoParseJson?: boolean;
    retryOptions?: {
        retries?: number;
        factor?: number;
        minTimeout?: number;
        maxTimeout?: number;
        randomize?: boolean;
        shouldRetry?: (err?: Error, req?: ServiceClientRequestOptions) => boolean;
        onRetry?: (currentAttempt?: number, err?: Error, req?: ServiceClientRequestOptions) => void;
    };
    circuitBreaker?: false | CircuitBreakerOptions | CircuitBreakerFactory;
    defaultRequestOptions?: Partial<ServiceClientRequestOptions>;
}
/**
 * A custom error returned in case something goes wrong.
 */
export declare abstract class ServiceClientError extends Error {
    response?: ServiceClientResponse | undefined;
    timings?: Timings;
    timingPhases?: TimingPhases;
    retryErrors: ServiceClientError[];
    /**
     * Use `instanceof` checks instead.
     * @deprecated since 0.9.0
     */
    type: string;
    protected constructor(originalError: Error, type: string, response?: ServiceClientResponse | undefined, name?: string);
}
export declare class CircuitOpenError extends ServiceClientError {
    constructor(originalError: Error, name: string);
}
export declare class BodyParseError extends ServiceClientError {
    response: ServiceClientResponse;
    constructor(originalError: Error, response: ServiceClientResponse, name: string);
}
export declare class RequestFilterError extends ServiceClientError {
    constructor(originalError: Error, name: string);
}
export declare class ResponseFilterError extends ServiceClientError {
    response: ServiceClientResponse;
    constructor(originalError: Error, response: ServiceClientResponse, name: string);
}
export declare class RequestNetworkError extends ServiceClientError {
    requestOptions: ServiceClientRequestOptions;
    constructor(originalError: RequestError, name: string);
}
export declare class RequestConnectionTimeoutError extends ServiceClientError {
    requestOptions: ServiceClientRequestOptions;
    constructor(originalError: RequestError, name: string);
}
export declare class RequestReadTimeoutError extends ServiceClientError {
    requestOptions: ServiceClientRequestOptions;
    constructor(originalError: RequestError, name: string);
}
export declare class RequestUserTimeoutError extends ServiceClientError {
    requestOptions: ServiceClientRequestOptions;
    constructor(originalError: RequestError, name: string);
}
export declare class RequestBodyStreamError extends ServiceClientError {
    requestOptions: ServiceClientRequestOptions;
    constructor(originalError: RequestError, name: string);
}
export declare class ShouldRetryRejectedError extends ServiceClientError {
    constructor(originalError: Error, type: string, name: string);
}
export declare class MaximumRetriesReachedError extends ServiceClientError {
    constructor(originalError: Error, type: string, name: string);
}
export declare class InternalError extends ServiceClientError {
    constructor(originalError: Error, name: string);
}
export declare class ServiceClient {
    /**
     * This filter will mark 5xx responses as failures. This is relevant for the circuit breaker.
     */
    static treat5xxAsError: ServiceClientRequestFilter;
    /**
     * This filter will mark 4xx responses as failures. This is relevant for the circuit breaker.
     *
     * This is not the default behaviour!
     */
    static treat4xxAsError: ServiceClientRequestFilter;
    /**
     * Use `instanceof BodyParseError` check instead
     * @deprecated since 0.9.0
     */
    static BODY_PARSE_FAILED: string;
    /**
     * Use `instanceof RequestFailedError` check instead
     * @deprecated since 0.9.0
     */
    static REQUEST_FAILED: string;
    /**
     * Use `instanceof RequestFilterError` check instead
     * @deprecated since 0.9.0
     */
    static REQUEST_FILTER_FAILED: string;
    /**
     * Use `instanceof ResponseFilterError` check instead
     * @deprecated since 0.9.0
     */
    static RESPONSE_FILTER_FAILED: string;
    /**
     * Use `instanceof CircuitOpenError` check instead
     * @deprecated since 0.9.0
     */
    static CIRCUIT_OPEN: string;
    /**
     * Use `instanceof CircuitOpenError` check instead
     * @deprecated since 0.9.0
     */
    static INTERNAL_ERROR: string;
    /**
     * Default list of post-filters which includes
     * `ServiceClient.treat5xxAsError`
     */
    static DEFAULT_FILTERS: ReadonlyArray<ServiceClientRequestFilter>;
    /**
     * Interface that the response will implement.
     * Deprecated: Use ServiceClientResponse instead.
     * @see ServiceClientResponse
     * @deprecated
     */
    static Response: typeof ServiceClientResponse;
    /**
     * Interface, that errors will implement.
     * Deprecated: Use ServiceClientError instead.
     * @see ServiceClientError
     * @deprecated
     */
    static Error: typeof ServiceClientError;
    name: string;
    private breaker?;
    private breakerFactory?;
    private options;
    /**
     * A ServiceClient can be constructed with all defaults by simply providing a URL, that can be parsed
     * by nodes url parser. Alternatively, provide actual ServiceClientOptions, that implement the
     * @{link ServiceClientOptions} interface.
     */
    constructor(optionsOrUrl: ServiceClientOptions | string);
    /**
     * Return an instance of a CircuitBreaker based on params.
     * Choses between a factory and a single static breaker
     */
    getCircuitBreaker(params: ServiceClientRequestOptions): CircuitBreakerPublicApi;
    /**
     * Perform a request to the service using given @{link ServiceClientRequestOptions}, returning the result in a promise.
     */
    request(userParams: ServiceClientRequestOptions): Promise<ServiceClientResponse>;
}
export default ServiceClient;
