"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const circuit_breaker_1 = require("./circuit-breaker");
exports.CircuitBreaker = circuit_breaker_1.CircuitBreaker;
const retry_1 = require("./retry");
const url = require("url");
const request_1 = require("./request");
exports.ServiceClientResponse = request_1.ServiceClientResponse;
/**
 * This interface describes all the options that may be passed to the service client at construction time.
 */
class ServiceClientOptions {
    constructor() {
        /**
         * This is the only mandatory option when creating a service client. All other properties have
         * reasonable defaults.
         */
        this.hostname = "";
    }
}
exports.ServiceClientOptions = ServiceClientOptions;
/**
 * Internal only, this interface guarantees, that the service client has all options available at runtime.
 */
class ServiceClientStrictOptions {
    constructor(options) {
        if (!options.hostname) {
            throw new Error("Please provide a `hostname` for this client");
        }
        this.hostname = options.hostname;
        this.filters = Array.isArray(options.filters)
            ? options.filters
            : [...ServiceClient.DEFAULT_FILTERS];
        this.timing = Boolean(options.timing);
        const autoParseJson = options.autoParseJson;
        this.autoParseJson = autoParseJson === undefined ? true : autoParseJson;
        this.retryOptions = Object.assign({ factor: 2, maxTimeout: 400, minTimeout: 200, randomize: true, retries: 0, shouldRetry() {
                return true;
            },
            onRetry() {
                /* do nothing */
            } }, options.retryOptions);
        if ((this.retryOptions.minTimeout || 0) > (this.retryOptions.maxTimeout || 0)) {
            throw new TypeError("The `maxTimeout` must be equal to or greater than the `minTimeout`");
        }
        this.defaultRequestOptions = Object.assign({ pathname: "/", protocol: "https:", timeout: 2000 }, options.defaultRequestOptions);
    }
}
/**
 * A custom error returned in case something goes wrong.
 */
class ServiceClientError extends Error {
    constructor(originalError, type, response, name = "ServiceClient") {
        super(`${name}: ${type}. ${originalError.message || ""}`);
        this.response = response;
        this.type = type;
        this.retryErrors = [];
        // Does not copy `message` from the original error
        Object.assign(this, originalError);
        const timingSource = response || originalError;
        this.timings = timingSource.timings;
        this.timingPhases = timingSource.timingPhases;
    }
}
exports.ServiceClientError = ServiceClientError;
class CircuitOpenError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.CIRCUIT_OPEN, undefined, name);
    }
}
exports.CircuitOpenError = CircuitOpenError;
class BodyParseError extends ServiceClientError {
    constructor(originalError, response, name) {
        super(originalError, ServiceClient.BODY_PARSE_FAILED, response, name);
        this.response = response;
    }
}
exports.BodyParseError = BodyParseError;
class RequestFilterError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.REQUEST_FILTER_FAILED, undefined, name);
    }
}
exports.RequestFilterError = RequestFilterError;
class ResponseFilterError extends ServiceClientError {
    constructor(originalError, response, name) {
        super(originalError, ServiceClient.RESPONSE_FILTER_FAILED, response, name);
        this.response = response;
    }
}
exports.ResponseFilterError = ResponseFilterError;
class RequestNetworkError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.REQUEST_FAILED, undefined, name);
        this.requestOptions = originalError.requestOptions;
    }
}
exports.RequestNetworkError = RequestNetworkError;
class RequestConnectionTimeoutError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.REQUEST_FAILED, undefined, name);
        this.requestOptions = originalError.requestOptions;
    }
}
exports.RequestConnectionTimeoutError = RequestConnectionTimeoutError;
class RequestReadTimeoutError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.REQUEST_FAILED, undefined, name);
        this.requestOptions = originalError.requestOptions;
    }
}
exports.RequestReadTimeoutError = RequestReadTimeoutError;
class RequestUserTimeoutError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.REQUEST_FAILED, undefined, name);
        this.requestOptions = originalError.requestOptions;
    }
}
exports.RequestUserTimeoutError = RequestUserTimeoutError;
class RequestBodyStreamError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.REQUEST_FAILED, undefined, name);
        this.requestOptions = originalError.requestOptions;
    }
}
exports.RequestBodyStreamError = RequestBodyStreamError;
class ShouldRetryRejectedError extends ServiceClientError {
    constructor(originalError, type, name) {
        super(originalError, type, undefined, name);
    }
}
exports.ShouldRetryRejectedError = ShouldRetryRejectedError;
class MaximumRetriesReachedError extends ServiceClientError {
    constructor(originalError, type, name) {
        super(originalError, type, undefined, name);
    }
}
exports.MaximumRetriesReachedError = MaximumRetriesReachedError;
class InternalError extends ServiceClientError {
    constructor(originalError, name) {
        super(originalError, ServiceClient.INTERNAL_ERROR, undefined, name);
    }
}
exports.InternalError = InternalError;
const JSON_CONTENT_TYPE_REGEX = /application\/(.*?[+])?json/i;
/**
 * This function takes a response and if it is of type json, tries to parse the body.
 */
const decodeResponse = (client, response) => {
    const contentType = response.headers["content-type"] ||
        (response.body ? "application/json" : "");
    if (typeof response.body === "string" &&
        JSON_CONTENT_TYPE_REGEX.test(contentType)) {
        try {
            response.body = JSON.parse(response.body);
        }
        catch (error) {
            throw new BodyParseError(error, response, client.name);
        }
    }
    return response;
};
/**
 * Reducer function to unwind response filters.
 */
const unwindResponseFilters = (promise, filter) => {
    return promise.then(params => filter.response ? filter.response(params) : params);
};
/**
 * Actually performs the request and applies the available filters in their respective phases.
 */
const requestWithFilters = (client, requestOptions, filters, autoParseJson) => {
    const pendingResponseFilters = [];
    const requestFilterPromise = filters.reduce((promise, filter) => {
        return promise.then(params => {
            if (params instanceof request_1.ServiceClientResponse) {
                return params;
            }
            const filtered = filter.request ? filter.request(params) : params;
            // also apply this filter when unwinding the chain
            pendingResponseFilters.unshift(filter);
            return filtered;
        });
    }, Promise.resolve(requestOptions));
    return requestFilterPromise
        .catch((error) => {
        throw new RequestFilterError(error, client.name);
    })
        .then(paramsOrResponse => paramsOrResponse instanceof request_1.ServiceClientResponse
        ? paramsOrResponse
        : request_1.request(paramsOrResponse).catch((error) => {
            if (error instanceof request_1.ConnectionTimeoutError) {
                throw new RequestConnectionTimeoutError(error, client.name);
            }
            else if (error instanceof request_1.UserTimeoutError) {
                throw new RequestUserTimeoutError(error, client.name);
            }
            else if (error instanceof request_1.BodyStreamError) {
                throw new RequestBodyStreamError(error, client.name);
            }
            else if (error instanceof request_1.ReadTimeoutError) {
                throw new RequestReadTimeoutError(error, client.name);
            }
            else if (error instanceof request_1.NetworkError) {
                throw new RequestNetworkError(error, client.name);
            }
            else {
                throw error;
            }
        }))
        .then(rawResponse => autoParseJson ? decodeResponse(client, rawResponse) : rawResponse)
        .then(resp => pendingResponseFilters
        .reduce(unwindResponseFilters, Promise.resolve(resp))
        .catch(error => {
        throw new ResponseFilterError(error, resp, client.name);
    }));
};
const noop = () => {
    /* do nothing */
};
const noopBreaker = {
    run(command) {
        command(noop, noop);
    },
    forceClose: () => null,
    forceOpen: () => null,
    unforce: () => null,
    isOpen: () => false
};
const buildStatusCodeFilter = (isError) => {
    return {
        response(response) {
            if (isError(response.statusCode)) {
                return Promise.reject(new Error(`Response status ${response.statusCode}`));
            }
            return Promise.resolve(response);
        }
    };
};
class ServiceClient {
    /**
     * A ServiceClient can be constructed with all defaults by simply providing a URL, that can be parsed
     * by nodes url parser. Alternatively, provide actual ServiceClientOptions, that implement the
     * @{link ServiceClientOptions} interface.
     */
    constructor(optionsOrUrl) {
        let options;
        if (typeof optionsOrUrl === "string") {
            const { port, protocol, query, hostname = "", pathname = "/" } = url.parse(optionsOrUrl, true);
            options = {
                hostname,
                defaultRequestOptions: {
                    port,
                    protocol,
                    query,
                    // pathname will be overwritten in actual usage, we just guarantee a sane default
                    pathname
                }
            };
        }
        else {
            options = optionsOrUrl;
        }
        if (typeof options.circuitBreaker === "object") {
            const breakerOptions = Object.assign({ windowDuration: 10000, numBuckets: 10, timeoutDuration: 2000, errorThreshold: 50, volumeThreshold: 10 }, options.circuitBreaker);
            this.breaker = new circuit_breaker_1.CircuitBreaker(breakerOptions);
        }
        if (typeof options.circuitBreaker === "function") {
            this.breakerFactory = options.circuitBreaker;
        }
        this.options = new ServiceClientStrictOptions(options);
        this.name = options.name || options.hostname;
    }
    /**
     * Return an instance of a CircuitBreaker based on params.
     * Choses between a factory and a single static breaker
     */
    getCircuitBreaker(params) {
        if (this.breaker) {
            return this.breaker;
        }
        if (this.breakerFactory) {
            return this.breakerFactory(params);
        }
        return noopBreaker;
    }
    /**
     * Perform a request to the service using given @{link ServiceClientRequestOptions}, returning the result in a promise.
     */
    request(userParams) {
        const params = Object.assign({}, this.options.defaultRequestOptions, userParams);
        params.hostname = this.options.hostname;
        params.port = params.port || (params.protocol === "https:" ? 443 : 80);
        params.timing =
            params.timing !== undefined ? params.timing : this.options.timing;
        params.headers = Object.assign({ accept: "application/json" }, params.headers);
        const { retries, factor, minTimeout, maxTimeout, randomize, shouldRetry, onRetry } = this.options.retryOptions;
        const opts = {
            retries,
            factor,
            minTimeout,
            maxTimeout,
            randomize
        };
        const retryErrors = [];
        return new Promise((resolve, reject) => {
            const breaker = this.getCircuitBreaker(params);
            const retryOperation = retry_1.operation(opts, (currentAttempt) => {
                breaker.run((success, failure) => {
                    return requestWithFilters(this, params, this.options.filters || [], this.options.autoParseJson)
                        .then((result) => {
                        success();
                        result.retryErrors = retryErrors;
                        resolve(result);
                    })
                        .catch((error) => {
                        retryErrors.push(error);
                        failure();
                        if (!shouldRetry(error, params)) {
                            reject(new ShouldRetryRejectedError(error, error.type, this.name));
                            return;
                        }
                        if (!retryOperation.retry()) {
                            // Wrapping error when user does not want retries would result
                            // in bad developer experience where you always have to unwrap it
                            // knowing there is only one error inside, so we do not do that.
                            if (retries === 0) {
                                reject(error);
                            }
                            else {
                                reject(new MaximumRetriesReachedError(error, error.type, this.name));
                            }
                            return;
                        }
                        onRetry(currentAttempt + 1, error, params);
                    });
                }, () => {
                    reject(new CircuitOpenError(new Error(), this.name));
                });
            });
            retryOperation.attempt();
        }).catch((error) => {
            const rawError = error instanceof Error ? error : new Error(String(error));
            const wrappedError = rawError instanceof ServiceClientError
                ? rawError
                : new InternalError(rawError, this.name);
            wrappedError.retryErrors = retryErrors;
            throw wrappedError;
        });
    }
}
/**
 * This filter will mark 5xx responses as failures. This is relevant for the circuit breaker.
 */
ServiceClient.treat5xxAsError = buildStatusCodeFilter(statusCode => statusCode >= 500);
/**
 * This filter will mark 4xx responses as failures. This is relevant for the circuit breaker.
 *
 * This is not the default behaviour!
 */
ServiceClient.treat4xxAsError = buildStatusCodeFilter(statusCode => statusCode >= 400 && statusCode < 500);
/**
 * Use `instanceof BodyParseError` check instead
 * @deprecated since 0.9.0
 */
ServiceClient.BODY_PARSE_FAILED = "Parsing of the response body failed";
/**
 * Use `instanceof RequestFailedError` check instead
 * @deprecated since 0.9.0
 */
ServiceClient.REQUEST_FAILED = "HTTP Request failed";
/**
 * Use `instanceof RequestFilterError` check instead
 * @deprecated since 0.9.0
 */
ServiceClient.REQUEST_FILTER_FAILED = "Request filter marked request as failed";
/**
 * Use `instanceof ResponseFilterError` check instead
 * @deprecated since 0.9.0
 */
ServiceClient.RESPONSE_FILTER_FAILED = "Response filter marked request as failed";
/**
 * Use `instanceof CircuitOpenError` check instead
 * @deprecated since 0.9.0
 */
ServiceClient.CIRCUIT_OPEN = "Circuit breaker is open and prevented the request";
/**
 * Use `instanceof CircuitOpenError` check instead
 * @deprecated since 0.9.0
 */
ServiceClient.INTERNAL_ERROR = "Perron internal error due to a bug or misconfiguration";
/**
 * Default list of post-filters which includes
 * `ServiceClient.treat5xxAsError`
 */
ServiceClient.DEFAULT_FILTERS = Object.freeze([ServiceClient.treat5xxAsError]);
/**
 * Interface that the response will implement.
 * Deprecated: Use ServiceClientResponse instead.
 * @see ServiceClientResponse
 * @deprecated
 */
ServiceClient.Response = request_1.ServiceClientResponse;
/**
 * Interface, that errors will implement.
 * Deprecated: Use ServiceClientError instead.
 * @see ServiceClientError
 * @deprecated
 */
ServiceClient.Error = ServiceClientError;
exports.ServiceClient = ServiceClient;
Object.freeze(ServiceClient);
exports.default = ServiceClient;
//# sourceMappingURL=client.js.map