/// <reference types="node" />
import { IncomingHttpHeaders } from "http";
import { RequestOptions } from "https";
import { ServiceClientError } from "./client";
export declare const enum EventSource {
    HTTP_RESPONSE = "http_response",
    HTTP_REQUEST = "http_request",
    HTTP_RESPONSE_BODY_STREAM = "http_response_body_stream",
    SOCKET = "socket"
}
export declare const enum EventName {
    START = "start",
    END = "end",
    DNS = "dns",
    TLS = "tls",
    TIMEOUT = "timeout",
    ERROR = "error",
    BYTES = "bytes"
}
/**
 * The NodeJS typescript definitions has OutgoingHttpHeaders
 * with `undefined` as one of the possible values, but NodeJS
 * runtime will throw an error for undefined values in headers.
 *
 * This overwrites the headers type in the RequestOptions
 * and removes undefined from one of the possible values of headers.
 */
export interface OutgoingHttpHeaders {
    [header: string]: number | string | string[];
}
interface Span {
    log(keyValuePairs: {
        [key: string]: any;
    }, timestamp?: number): this;
}
export interface ServiceClientRequestOptions extends RequestOptions {
    pathname: string;
    query?: object;
    timing?: boolean;
    autoDecodeUtf8?: boolean;
    dropRequestAfter?: number;
    body?: any;
    headers?: OutgoingHttpHeaders;
    /**
     * Happens when the socket connection cannot be established
     */
    timeout?: number;
    /**
     * Happens after the socket connection is successfully established
     * and there is no activity on that socket
     */
    readTimeout?: number;
    /**
     * Opentracing like span interface to log events
     */
    span?: Span;
}
export declare class ServiceClientResponse {
    statusCode: number;
    headers: IncomingHttpHeaders;
    body: Buffer | string | object;
    request: ServiceClientRequestOptions;
    timings?: Timings;
    timingPhases?: TimingPhases;
    retryErrors: ServiceClientError[];
    constructor(statusCode: number, headers: IncomingHttpHeaders, body: Buffer | string | object, request: ServiceClientRequestOptions);
}
export interface Timings {
    lookup?: number;
    socket?: number;
    connect?: number;
    secureConnect?: number;
    response?: number;
    end?: number;
}
export interface TimingPhases {
    wait?: number;
    dns?: number;
    tcp?: number;
    tls?: number;
    firstByte?: number;
    download?: number;
    total?: number;
}
export declare class RequestError extends Error {
    timings?: Timings;
    timingPhases?: TimingPhases;
    requestOptions: ServiceClientRequestOptions;
    constructor(message: string, requestOptions: ServiceClientRequestOptions, timings?: Timings);
}
export declare class NetworkError extends RequestError {
    constructor(originalError: Error, requestOptions: ServiceClientRequestOptions, timings?: Timings);
}
export declare class ConnectionTimeoutError extends RequestError {
    constructor(requestOptions: ServiceClientRequestOptions, timings?: Timings);
}
export declare class ReadTimeoutError extends RequestError {
    constructor(requestOptions: ServiceClientRequestOptions, timings?: Timings);
}
export declare class UserTimeoutError extends RequestError {
    constructor(requestOptions: ServiceClientRequestOptions, timings?: Timings);
}
export declare class BodyStreamError extends RequestError {
    constructor(originalError: Error, requestOptions: ServiceClientRequestOptions, timings?: Timings);
}
export declare const request: (options: ServiceClientRequestOptions) => Promise<ServiceClientResponse>;
export {};
