"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const http_1 = require("http");
const https_1 = require("https");
const querystring = require("querystring");
const zlib = require("zlib");
const DEFAULT_READ_TIMEOUT = 2000;
const DEFAULT_CONNECTION_TIMEOUT = 1000;
const getInterval = (time) => {
    const diff = process.hrtime(time);
    return Math.round(diff[0] * 1000 + diff[1] / 1000000);
};
class ServiceClientResponse {
    constructor(statusCode, headers, body, request) {
        this.statusCode = statusCode;
        this.headers = headers;
        this.body = body;
        this.request = request;
        this.retryErrors = [];
    }
}
exports.ServiceClientResponse = ServiceClientResponse;
const subtract = (a, b) => {
    if (typeof a === "number" && typeof b === "number") {
        return a - b;
    }
    return undefined;
};
class RequestError extends Error {
    constructor(message, requestOptions, timings) {
        super(message);
        this.requestOptions = requestOptions;
        this.timings = timings;
        this.timingPhases = timings && makeTimingPhases(timings);
    }
}
exports.RequestError = RequestError;
class NetworkError extends RequestError {
    constructor(originalError, requestOptions, timings) {
        super(originalError.message, requestOptions, timings);
        this.stack = originalError.stack;
    }
}
exports.NetworkError = NetworkError;
class ConnectionTimeoutError extends RequestError {
    constructor(requestOptions, timings) {
        super("socket timeout", requestOptions, timings);
    }
}
exports.ConnectionTimeoutError = ConnectionTimeoutError;
class ReadTimeoutError extends RequestError {
    constructor(requestOptions, timings) {
        super("read timeout", requestOptions, timings);
    }
}
exports.ReadTimeoutError = ReadTimeoutError;
class UserTimeoutError extends RequestError {
    constructor(requestOptions, timings) {
        super("request timeout", requestOptions, timings);
    }
}
exports.UserTimeoutError = UserTimeoutError;
class BodyStreamError extends RequestError {
    constructor(originalError, requestOptions, timings) {
        super(originalError.message, requestOptions, timings);
        this.stack = originalError.stack;
    }
}
exports.BodyStreamError = BodyStreamError;
const makeTimingPhases = (timings) => {
    return {
        wait: timings.socket,
        dns: subtract(timings.lookup, timings.socket),
        tcp: subtract(timings.connect, timings.lookup),
        tls: subtract(timings.secureConnect, timings.connect),
        firstByte: subtract(timings.response, timings.secureConnect),
        download: subtract(timings.end, timings.response),
        total: timings.end
    };
};
exports.request = (options) => {
    options = Object.assign({ protocol: "https:", autoDecodeUtf8: true }, options);
    function logEvent(source, name, value) {
        if (options.span != null) {
            options.span.log({
                [source]: value != null ? { name, value } : name
            });
        }
    }
    if ("pathname" in options && !("path" in options)) {
        if ("query" in options) {
            let query = querystring.stringify(options.query);
            if (query) {
                query = "?" + query;
            }
            options.path = `${options.pathname}${query}`;
        }
        else {
            options.path = options.pathname;
        }
    }
    const connectionTimeout = options.timeout || DEFAULT_CONNECTION_TIMEOUT;
    const readTimeout = options.readTimeout || DEFAULT_READ_TIMEOUT;
    const httpRequestFn = options.protocol === "https:" ? https_1.request : http_1.request;
    return new Promise((resolve, reject) => {
        let hasRequestEnded = false;
        let startTime;
        let timings;
        if (options.timing) {
            startTime = process.hrtime();
            timings = {
                lookup: undefined,
                socket: undefined,
                connect: undefined,
                secureConnect: undefined,
                response: undefined,
                end: undefined
            };
        }
        const requestObject = httpRequestFn(options);
        requestObject.setTimeout(readTimeout, () => {
            logEvent("http_request" /* HTTP_REQUEST */, "timeout" /* TIMEOUT */);
            requestObject.socket.destroy();
            reject(new ReadTimeoutError(options));
        });
        requestObject.once("error", err => {
            hasRequestEnded = true;
            logEvent("http_request" /* HTTP_REQUEST */, "error" /* ERROR */, err.message);
            reject(new NetworkError(err, options));
        });
        // Fires once the socket is assigned to a request
        requestObject.once("socket", (socket) => {
            logEvent("socket" /* SOCKET */, "start" /* START */);
            if (options.timing) {
                timings.socket = getInterval(startTime);
            }
            if (socket.connecting) {
                socket.setTimeout(connectionTimeout, () => {
                    logEvent("socket" /* SOCKET */, "timeout" /* TIMEOUT */);
                    // socket should be manually cleaned up
                    socket.destroy();
                    reject(new ConnectionTimeoutError(options));
                });
                socket.once("lookup", () => {
                    logEvent("socket" /* SOCKET */, "dns" /* DNS */);
                    if (options.timing) {
                        timings.lookup = getInterval(startTime);
                    }
                });
                // connect event would kick in only for new socket connections
                // and not for connections that are kept alive
                socket.once("connect", () => {
                    logEvent("socket" /* SOCKET */, "end" /* END */);
                    if (options.timing) {
                        timings.connect = getInterval(startTime);
                    }
                });
                socket.once("secureConnect", () => {
                    logEvent("http_request" /* HTTP_REQUEST */, "tls" /* TLS */);
                    if (options.timing) {
                        timings.secureConnect = getInterval(startTime);
                    }
                });
            }
            else {
                if (options.timing) {
                    timings.lookup = timings.socket;
                    timings.connect = timings.socket;
                    timings.secureConnect = timings.socket;
                }
            }
        });
        requestObject.on("response", (response) => {
            logEvent("http_response" /* HTTP_RESPONSE */, "start" /* START */);
            if (options.timing) {
                timings.response = getInterval(startTime);
            }
            const { headers, statusCode } = response;
            let bodyStream;
            const encoding = headers && headers["content-encoding"];
            if (encoding === "gzip" || encoding === "deflate") {
                response.on("error", err => {
                    logEvent("http_response" /* HTTP_RESPONSE */, "error" /* ERROR */, err.message);
                    reject(new NetworkError(err, options));
                });
                bodyStream = response.pipe(zlib.createUnzip());
            }
            else {
                bodyStream = response;
            }
            let chunks = [];
            let bufferLength = 0;
            bodyStream.on("error", err => {
                logEvent("http_response_body_stream" /* HTTP_RESPONSE_BODY_STREAM */, "error" /* ERROR */, err.message);
                reject(new NetworkError(err, options));
            });
            bodyStream.on("data", data => {
                logEvent("http_response_body_stream" /* HTTP_RESPONSE_BODY_STREAM */, "bytes" /* BYTES */, data.length);
                bufferLength += data.length;
                chunks.push(data);
            });
            bodyStream.on("end", () => {
                logEvent("http_response_body_stream" /* HTTP_RESPONSE_BODY_STREAM */, "end" /* END */, bufferLength);
                hasRequestEnded = true;
                let body;
                const bufferedBody = Buffer.concat(chunks, bufferLength);
                if (options.autoDecodeUtf8) {
                    body = bufferedBody.toString("utf8");
                }
                else {
                    body = bufferedBody;
                }
                // to avoid leaky behavior
                chunks = [];
                bufferLength = 0;
                const serviceClientResponse = new ServiceClientResponse(statusCode || 0, headers, body, options);
                if (options.timing) {
                    timings.end = getInterval(startTime);
                    serviceClientResponse.timings = timings;
                    serviceClientResponse.timingPhases = makeTimingPhases(timings);
                }
                resolve(serviceClientResponse);
                logEvent("http_response" /* HTTP_RESPONSE */, "end" /* END */);
            });
        });
        if (options.dropRequestAfter) {
            setTimeout(() => {
                if (!hasRequestEnded) {
                    requestObject.abort();
                    const err = new UserTimeoutError(options, timings);
                    logEvent("http_request" /* HTTP_REQUEST */, "error" /* ERROR */, err.message);
                    reject(err);
                }
            }, options.dropRequestAfter);
        }
        logEvent("http_request" /* HTTP_REQUEST */, "start" /* START */);
        if (options.body) {
            if (typeof options.body.pipe === "function") {
                const requestBody = options.body;
                requestBody.pipe(requestObject);
                requestBody.on("error", err => {
                    requestObject.abort();
                    reject(new BodyStreamError(err, options, timings));
                });
                return;
            }
            requestObject.write(options.body);
        }
        requestObject.end();
    });
};
//# sourceMappingURL=request.js.map